// Simple Hall-effect magnetometer.
// Version 1.0
// A. Kruger
// Iowa City, Iowa 2016
//
// Uses Arduno Nano

#include <LiquidCrystal.h>

const int numRows = 2;            // LCD size
const int numCols = 16;           // LCD size
LiquidCrystal lcd(12, 11, 5, 4, 3, 2); // LCD object + pin numbers, marked D12, D11, D5,..., D2 on Nano.


#define NO_REF 0            // Don't subtrack reference voltage.
#define REF    1            // Subtract reference voltage.

bool state = NO_REF;        // Keeps track of which display mode is active.

double hallVoltage(int n);

double ref = 0.0;
char data[100];

void setup() {
  pinMode(6, INPUT);               // Display mode switch is connected to pin 6
  Serial.begin(9600);
  Serial1.begin(9600);
  lcd.begin(numCols, numRows);
  lcd.print("Magnetometer 1.0");

  // When we start out, display the raw Hall-effect sensor voltage.

  ref = 0.0;
  state = NO_REF;

  // Make the ADC use the reference on the AREF pin. This is a 4.096 V (nominal)
  // voltage reference.

  analogReference(EXTERNAL);
}

void loop() {
  double ave;
  int i, n;
  String str;
  char mode = 'A';

  // Get the average of "n" samples of the Hall-effect sensor.
  // Each sample takes about 0.173 ms

  n = 2900 / 2;                   // About 0.25 seconds
  ave = hallVoltage(n);
  ave = ave * (4.096 / 1023.0);
  ave = ave - ref;

  // We display to 3 decimal points. The next line makes sure the display
  // does not dither between +0.000 and -0.000.

  if (abs(ave) < 0.001) ave = abs(ave);

  str =  "Val = " + String(ave, 3) + " V";
  Serial.println(str);

  // The following erases (by writing blanks) part of the display
  // and then write the new text.  This ensures a smooth display.

  lcd.setCursor(7, 1);          // 1 => second row
  lcd.print("        ");        // 8 blanks
  lcd.setCursor(0, 1);          // Back to start of line
  lcd.print(str);

  // If user pressed mode button, toggle display mode.

  if (digitalRead(6) == LOW) {
    switch (state) {
      case NO_REF:
        ref = hallVoltage(n);
        ref = ref * (4.096 / 1023.0);
        state = REF;
        break;
      case REF:
        state = NO_REF;
        ref = 0;
        break;
      default:
        state = NO_REF;
    }
    while (digitalRead(6) == LOW)
      delay(10); // Debounce
  }
  // Communication with RX0
  if (Serial1.peek() != -1) {
    c = Serial1.read();
    switch (c) {
      case M:                   // Mode Switch
        if (mode == 'A')        // Automatic -> Semi-Automatic
          mode = 'S';
        else if (mode == 'S')   // Semi-Automatic -> Automatic
          mode = 'A';
        break;
      case A:                   // Force Automatic
        mode = 'A';
        break;
      case S:                   // Semi-Automatic
        if (mode == 'S')        // Print Measurement
          Serial1.println(str);
        else
          mode = 'S';           // Force Semi-Automatic
        break;
    }
  }
  else if (Serial1.peek() == -1 and mode == 'A') // No Change Needed
    Serial1.println(str);
}

// Returns the mean of "n" samples of the Hall-effect sensor
// Use an iterative algorithm to calculate the mean.
// See: http://www-uxsup.csx.cam.ac.uk/~fanf2/hermes/doc/antiforgery/stats.pdf
double hallVoltage(int n)
{
  double u, u1; // Mean (u) and previous mean (u1)
  double x;     // Current data value
  int i;

  // Make a few "init" measurements.

  x = (double)analogRead(A0);
  x = (double)analogRead(A0);
  x = (double)analogRead(A0);

  x = (double)analogRead(A0);
  u = x;
  for (i = 1; i <= n - 1; i++) {
    u1 = u;
    u = u1 + (x - u1) / i;
    x = (double)analogRead(A0);
  }
  return (u);
}
